@@ -3,8 +3,8 @@
  * directory.
  */
 
+#define LIBISOFS_WITHOUT_LIBBURN yes
 #include "libisofs.h"
-#include "libburn/libburn.h"
 
 #include <stdio.h>
 #include <getopt.h>
