@@ -1,19 +1,27 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
+ * Copyright (c) 2009 Thomas Schmitt
  * 
  * This file is part of the libisofs project; you can redistribute it and/or 
  * modify it under the terms of the GNU General Public License version 2 as 
  * published by the Free Software Foundation. See COPYING file for details.
  */
 
+/* libisofs.h defines aaip_xinfo_func */
+#include "libisofs.h"
+
 #include "builder.h"
 #include "node.h"
 #include "fsource.h"
+#include "image.h"
+#include "aaip_0_2.h"
 
 #include <stdlib.h>
 #include <string.h>
 #include <limits.h>
 
+
+
 void iso_node_builder_ref(IsoNodeBuilder *builder)
 {
     ++builder->refcount;
@@ -85,6 +93,8 @@ int default_create_node(IsoNodeBuilder *builder, IsoImage *image,
     struct stat info;
     IsoNode *new;
     char *name;
+    unsigned char *aa_string;
+    char *a_text = NULL, *d_text = NULL;
 
     if (builder == NULL || src == NULL || node == NULL) {
         return ISO_NULL_POINTER;
@@ -166,7 +176,7 @@ int default_create_node(IsoNodeBuilder *builder, IsoImage *image,
     }
 
     /* fill fields */
-    iso_node_set_permissions(new, info.st_mode);
+    iso_node_set_perms_internal(new, info.st_mode, 1);
     iso_node_set_uid(new, info.st_uid);
     iso_node_set_gid(new, info.st_gid);
     iso_node_set_atime(new, info.st_atime);
@@ -174,7 +184,31 @@ int default_create_node(IsoNodeBuilder *builder, IsoImage *image,
     iso_node_set_ctime(new, info.st_ctime);
     iso_node_set_uid(new, info.st_uid);
 
+    /* Eventually set S_IRWXG from ACL */
+    if (image->builder_ignore_acl) {
+        ret = iso_file_source_get_aa_string(src, &aa_string, 4);
+        if (aa_string != NULL)
+            iso_aa_get_acl_text(aa_string, info.st_mode, &a_text, &d_text, 16);
+        if (a_text != NULL) {
+            aaip_cleanout_st_mode(a_text, &(info.st_mode), 4 | 16);
+            iso_node_set_perms_internal(new, info.st_mode, 1);
+        }
+        iso_aa_get_acl_text(aa_string, info.st_mode, &a_text, &d_text,
+                            1 << 15); /* free ACL texts */
+    }
+
+    /* Obtain ownership of eventual AA string */
+    ret = iso_file_source_get_aa_string(src, &aa_string,
+            1 | (image->builder_ignore_acl << 1) |
+                (image->builder_ignore_ea << 2 ));
+    if (ret == 1 && aa_string != NULL) {
+        ret = iso_node_add_xinfo(new, aaip_xinfo_func, aa_string);
+        if (ret < 0)
+            return ret;
+    }
+
     *node = new;
+
     return ISO_SUCCESS;
 }
 
