@@ -1,13 +1,19 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
  * Copyright (c) 2007 Mario Danic
+ * Copyright (c) 2009 Thomas Schmitt
  *
  * This file is part of the libisofs project; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License version 2 as
  * published by the Free Software Foundation. See COPYING file for details.
  */
 
+/*
+   Use the copy of the struct burn_source definition in libisofs.h
+*/
+#define LIBISOFS_WITHOUT_LIBBURN yes
 #include "libisofs.h"
+
 #include "ecma119.h"
 #include "joliet.h"
 #include "iso1999.h"
@@ -21,8 +27,6 @@
 #include "util.h"
 #include "system_area.h"
 
-#include "libburn/libburn.h"
-
 #include <stdlib.h>
 #include <time.h>
 #include <string.h>
@@ -415,7 +419,11 @@ int ecma119_writer_write_vol_desc(IsoImageWriter *writer)
 }
 
 static
+#ifdef Libisofs_use_parent_for_dot_doT
+int write_one_dir(Ecma119Image *t, Ecma119Node *dir, Ecma119Node *parent)
+#else
 int write_one_dir(Ecma119Image *t, Ecma119Node *dir)
+#endif
 {
     int ret;
     uint8_t buffer[BLOCK_SIZE];
@@ -458,7 +466,11 @@ int write_one_dir(Ecma119Image *t, Ecma119Node *dir)
         }
     }
     len = 34 + info.suf_len;
+#ifdef Libisofs_use_parent_for_dot_doT
+    write_one_dir_record(t, parent, 1, buf, 1, &info, 0);
+#else
     write_one_dir_record(t, dir, 1, buf, 1, &info, 0);
+#endif
     buf += len;
 
     for (i = 0; i < dir->info.dir->nchildren; i++) {
@@ -515,13 +527,24 @@ int write_one_dir(Ecma119Image *t, Ecma119Node *dir)
 }
 
 static
+#ifdef Libisofs_use_parent_for_dot_doT
+
+/* ts A90309 : todo ? : rename parameter "root" to "dir" */
+
+int write_dirs(Ecma119Image *t, Ecma119Node *root, Ecma119Node *parent)
+#else
 int write_dirs(Ecma119Image *t, Ecma119Node *root)
+#endif
 {
     int ret;
     size_t i;
 
     /* write all directory entries for this dir */
+#ifdef Libisofs_use_parent_for_dot_doT
+    ret = write_one_dir(t, root, parent);
+#else
     ret = write_one_dir(t, root);
+#endif
     if (ret < 0) {
         return ret;
     }
@@ -530,7 +553,11 @@ int write_dirs(Ecma119Image *t, Ecma119Node *root)
     for (i = 0; i < root->info.dir->nchildren; i++) {
         Ecma119Node *child = root->info.dir->children[i];
         if (child->type == ECMA119_DIR) {
+#ifdef Libisofs_use_parent_for_dot_doT
+            ret = write_dirs(t, child, root);
+#else
             ret = write_dirs(t, child);
+#endif
             if (ret < 0) {
                 return ret;
             }
@@ -649,7 +676,11 @@ int ecma119_writer_write_data(IsoImageWriter *writer)
     t = writer->target;
 
     /* first of all, we write the directory structure */
+#ifdef Libisofs_use_parent_for_dot_doT
+    ret = write_dirs(t, t->root, t->root);
+#else
     ret = write_dirs(t, t->root);
+#endif
     if (ret < 0) {
         return ret;
     }
@@ -861,7 +892,7 @@ int ecma119_image_new(IsoImage *src, IsoWriteOpts *opts, Ecma119Image **img)
 {
     int ret, i, voldesc_size, nwriters;
     Ecma119Image *target;
-    int el_torito_writer_index = -1;
+    int el_torito_writer_index = -1, file_src_writer_index= -1;
 
     /* 1. Allocate target and copy opts there */
     target = calloc(1, sizeof(Ecma119Image));
@@ -883,6 +914,7 @@ int ecma119_image_new(IsoImage *src, IsoWriteOpts *opts, Ecma119Image **img)
     target->rockridge = opts->rockridge;
     target->joliet = opts->joliet;
     target->iso1999 = opts->iso1999;
+    target->aaip = opts->aaip;
     target->always_gmt = opts->always_gmt;
     target->ino = 0;
     target->omit_version_numbers = opts->omit_version_numbers
@@ -896,6 +928,7 @@ int ecma119_image_new(IsoImage *src, IsoWriteOpts *opts, Ecma119Image **img)
     target->relaxed_vol_atts = opts->relaxed_vol_atts;
     target->joliet_longer_paths = opts->joliet_longer_paths;
     target->rrip_version_1_10 = opts->rrip_version_1_10;
+    target->aaip_susp_1_10 = opts->aaip_susp_1_10;
     target->dir_rec_mtime = opts->dir_rec_mtime;
     target->sort_files = opts->sort_files;
 
@@ -1026,6 +1059,7 @@ int ecma119_image_new(IsoImage *src, IsoWriteOpts *opts, Ecma119Image **img)
     if (ret < 0) {
         goto target_cleanup;
     }
+    file_src_writer_index = target->nwriters - 1;
 
     /*
      * 3.
@@ -1042,6 +1076,12 @@ int ecma119_image_new(IsoImage *src, IsoWriteOpts *opts, Ecma119Image **img)
         if (i == el_torito_writer_index)
     continue;
 #endif
+
+        /* ts A90228 : exposing address of data start to IsoWriteOpts */
+        if (i == file_src_writer_index) {
+            opts->data_start_lba = target->curblock;
+        }
+
         ret = writer->compute_data_blocks(writer);
         if (ret < 0) {
             goto target_cleanup;
@@ -1412,6 +1452,15 @@ int iso_write_opts_set_iso1999(IsoWriteOpts *opts, int enable)
     return ISO_SUCCESS;
 }
 
+int iso_write_opts_set_aaip(IsoWriteOpts *opts, int enable)
+{
+    if (opts == NULL) {
+        return ISO_NULL_POINTER;
+    }
+    opts->aaip = enable ? 1 : 0;
+    return ISO_SUCCESS;
+}
+
 int iso_write_opts_set_omit_version_numbers(IsoWriteOpts *opts, int omit)
 {
     if (opts == NULL) {
@@ -1502,6 +1551,15 @@ int iso_write_opts_set_rrip_version_1_10(IsoWriteOpts *opts, int oldvers)
     return ISO_SUCCESS;
 }
 
+int iso_write_opts_set_aaip_susp_1_10(IsoWriteOpts *opts, int oldvers)
+{
+    if (opts == NULL) {
+        return ISO_NULL_POINTER;
+    }
+    opts->aaip_susp_1_10 = oldvers ? 1 : 0;
+    return ISO_SUCCESS;
+}
+
 int iso_write_opts_set_dir_rec_mtime(IsoWriteOpts *opts, int allow)
 {
     if (opts == NULL) {
@@ -1658,3 +1716,13 @@ int iso_write_opts_set_fifo_size(IsoWriteOpts *opts, size_t fifo_size)
     opts->fifo_size = fifo_size;
     return ISO_SUCCESS;
 }
+
+int iso_write_opts_get_data_start(IsoWriteOpts *opts, uint32_t *data_start,
+                                  int flag)
+{
+    if(opts->data_start_lba == 0)
+	return ISO_ERROR;
+    *data_start = opts->data_start_lba;
+    return ISO_SUCCESS;
+}
+
