@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
+ * Copyright (c) 2009 Thomas Schmitt
  *
  * This file is part of the libisofs project; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License version 2 as
@@ -42,6 +43,8 @@ struct iso_write_opts {
     unsigned int joliet :1;
     unsigned int iso1999 :1;
 
+    unsigned int aaip :1; /* whether to write eventual ACL and EAs */
+
     /* allways write timestamps in GMT */
     unsigned int always_gmt :1;
 
@@ -112,6 +115,19 @@ struct iso_write_opts {
      */
     unsigned int rrip_version_1_10 :1;
 
+    /**
+     * Write AAIP as extension according to SUSP 1.10 rather than SUSP 1.12.
+     * I.e. without announcing it by an ER field and thus without the need
+     * to preceed the RRIP fields by an ES and to preceed the AA field by ES.
+     * This saves bytes and might avoid problems with readers which dislike
+     * ER fields other than the ones for RRIP.
+     * On the other hand, SUSP 1.12 frowns on such unannounced extensions
+     * and prescribes ER and ES. It does this since year 1994.
+     *
+     * In effect only if above flag .aaip is set to 1.
+     */
+    unsigned int aaip_susp_1_10 :1;
+
     /**
      * Store as ECMA-119 Directory Record timestamp the mtime of the source
      * rather than the image creation time. (The ECMA-119 prescription seems
@@ -223,6 +239,16 @@ struct iso_write_opts {
      * buffer.
      */
     size_t fifo_size;
+
+
+    /* ts A90228 : exposing address of data start block to the API */
+    /**
+     * This is not an option setting but a value returned after the options
+     * were used to compute the layout of the image.
+     * It tells the LBA of the first plain file data block in the image.
+     */
+    uint32_t data_start_lba;
+
 };
 
 typedef struct ecma119_image Ecma119Image;
@@ -245,6 +271,8 @@ struct ecma119_image
     unsigned int eltorito :1;
     unsigned int iso1999 :1;
 
+    unsigned int aaip :1;  /* whether to write eventual ACLs and EAs */
+
     /* allways write timestamps in GMT */
     unsigned int always_gmt :1;
 
@@ -265,6 +293,9 @@ struct ecma119_image
     /** Write old fashioned RRIP-1.10 rather than RRIP-1.12 */
     unsigned int rrip_version_1_10 :1;
 
+    /* Write AAIP as extension according to SUSP 1.10 rather than SUSP 1.12. */
+    unsigned int aaip_susp_1_10 :1;
+
     /* Store in ECMA-119 timestamp mtime of source */
     unsigned int dir_rec_mtime :1;
 
@@ -507,4 +538,5 @@ struct ecma119_vol_desc_terminator
     uint8_t reserved                 BP(8, 2048);
 };
 
+
 #endif /*LIBISO_ECMA119_H_*/
