@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
+ * Copyright (c) 2009 Thomas Schmitt
  *
  * This file is part of the libisofs project; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License version 2 as
@@ -18,6 +19,7 @@
 #include "image.h"
 #include "tree.h"
 #include "eltorito.h"
+#include "aaip_0_2.h"
 
 #include <stdlib.h>
 #include <string.h>
@@ -57,6 +59,7 @@ struct iso_read_opts
     unsigned int norock : 1; /*< Do not read Rock Ridge extensions */
     unsigned int nojoliet : 1; /*< Do not read Joliet extensions */
     unsigned int noiso1999 : 1; /*< Do not read ISO 9660:1999 enhanced tree */
+    unsigned int noaaip : 1; /* Do not read AAIP extension for xattr and ACL */
 
     /**
      * When both Joliet and RR extensions are present, the RR tree is used.
@@ -213,6 +216,16 @@ typedef struct
     /** If ISO 9660:1999 is available on image */
     unsigned int iso1999 : 1;
 
+    /** Whether AAIP info shall be loaded if it is present.
+     *  1 = yes , 0 = no
+     */
+    int aaip_load;
+
+    /** Whether AAIP is present. Version major.minor = major * 100 + minor
+     *  Value -1 means that no AAIP ER was detected yet.
+     */
+    int aaip_version;
+
     /**
      * Number of blocks of the volume, as reported in the PVM.
      */
@@ -228,6 +241,10 @@ typedef struct
     uint32_t imgblock; /**< Block for El-Torito boot image */
     uint32_t catblock; /**< Block for El-Torito catalog */
 
+    /* ts A90303 */
+    /* Inode number generator counter */
+    ino_t inode_counter;
+
 } _ImageFsData;
 
 typedef struct image_fs_data ImageFileSourceData;
@@ -265,6 +282,14 @@ struct image_fs_data
          */
         off_t offset;
     } data;
+
+    /**
+     * malloc() storage for the string of AA fields which represent
+     * ACLs and XFS-style Extended Attributes. (Not to be confused with
+     * ECMA-119 Extended Attributes.)
+     */
+    unsigned char *aa_string;
+
 };
 
 struct child_list
@@ -907,11 +932,37 @@ void ifs_free(IsoFileSource *src)
 
     free(data->sections);
     free(data->name);
+    if (data->aa_string != NULL)
+        free(data->aa_string);
     free(data);
 }
 
+
+static
+int ifs_get_aa_string(IsoFileSource *src, unsigned char **aa_string, int flag)
+{
+    size_t len;
+    ImageFileSourceData *data;
+
+    data = src->data;
+
+    if ((flag & 1) || data->aa_string == NULL) {
+        *aa_string = data->aa_string;
+        data->aa_string = NULL;
+    } else {
+        len = aaip_count_bytes(data->aa_string, 0);
+        *aa_string = calloc(len, 1);
+        if (*aa_string == NULL)
+            return ISO_OUT_OF_MEM;
+        memcpy(*aa_string, data->aa_string, len);
+    }
+    return 1;
+}
+
+
 IsoFileSourceIface ifs_class = {
-    0, /* version */
+
+    1, /* version */
     ifs_get_path,
     ifs_get_name,
     ifs_lstat,
@@ -924,7 +975,9 @@ IsoFileSourceIface ifs_class = {
     ifs_readlink,
     ifs_get_filesystem,
     ifs_free,
-    ifs_lseek
+    ifs_lseek,
+    ifs_get_aa_string
+
 };
 
 /**
@@ -964,6 +1017,31 @@ char *get_name(_ImageFsData *fsdata, const char *str, size_t len)
     return name;
 }
 
+
+/* ts A90303 */
+/**
+ * A global counter for default inode numbers for the ISO image filesystem.
+ * @param fs    The filesystem where the number shall be used
+ * @param flag  bit0= reset count
+ */
+static
+ino_t fs_give_ino_number(IsoImageFilesystem *fs, int flag)
+{
+    _ImageFsData *fsdata;
+
+    fsdata = (_ImageFsData*)fs->data;
+    if (flag & 1)
+        fsdata->inode_counter = 0;
+    fsdata->inode_counter++;
+    if (fsdata->inode_counter == 0) {
+
+        /* >>> raise alert because of inode rollover */;
+     
+    }
+    return fsdata->inode_counter;
+}
+
+
 /**
  *
  * @param src
@@ -996,6 +1074,10 @@ int iso_file_source_new_ifs(IsoImageFilesystem *fs, IsoFileSource *parent,
 
     uint32_t relocated_dir = 0;
 
+    unsigned char *aa_string = NULL;
+    size_t aa_size = 0, aa_len = 0, prev_field = 0;
+    int aa_done = 0;
+
     if (fs == NULL || fs->data == NULL || record == NULL || src == NULL) {
         return ISO_NULL_POINTER;
     }
@@ -1025,7 +1107,7 @@ int iso_file_source_new_ifs(IsoImageFilesystem *fs, IsoFileSource *parent,
     if (record->len_xa[0]) {
         iso_msg_submit(fsdata->msgid, ISO_UNSUPPORTED_ECMA119, 0,
               "Unsupported image. This image has at least one file with "
-              "Extended Attributes, that are not supported");
+              "ECMA-119 Extended Attributes, that are not supported");
         return ISO_UNSUPPORTED_ECMA119;
     }
 
@@ -1235,9 +1317,28 @@ int iso_file_source_new_ifs(IsoImageFilesystem *fs, IsoFileSource *parent,
                                   "than '.' entry of root node");
                 }
                 continue;
+
+            /* Need to read AA in any case so it is available for S_IRWXG
+               mapping in case that fsdata->aaip_load != 1
+             */
+            } else if (SUSP_SIG(sue, 'A', 'A')) {
+
+                ret = read_aaip_AA(sue, &aa_string, &aa_size, &aa_len,
+                                   &prev_field, &aa_done, 0);
+                if (ret < 0) {
+                    /* notify and continue */
+                    ret = iso_msg_submit(fsdata->msgid, ISO_WRONG_RR_WARN, ret,
+                                 "Invalid AA entry");
+                    continue;
+                }
+
+/* This message is inflationary */
+/*
             } else {
                 ret = iso_msg_submit(fsdata->msgid, ISO_SUSP_UNHANDLED, 0,
                     "Unhandled SUSP entry %c%c.", sue->sig[0], sue->sig[1]);
+*/
+
             }
         }
 
@@ -1388,6 +1489,18 @@ int iso_file_source_new_ifs(IsoImageFilesystem *fs, IsoFileSource *parent,
         return ISO_SUCCESS;
     }
 
+
+#ifdef Libisofs_new_fs_image_inO
+
+    if (fsdata->rr != RR_EXT_112) {
+        if (fsdata->rr == 0) {
+            atts.st_nlink = 1;
+        }
+    }
+    atts.st_ino = fs_give_ino_number(fs, 0);
+
+#else /* Libisofs_new_fs_image_inO */
+
     /* ts Nov 25 2008: TODO
        This seems not fully consistent with read_rr_PX() which decides
        by (px->len_sue[0] == 44) whether an inode number is present or not.
@@ -1408,16 +1521,13 @@ int iso_file_source_new_ifs(IsoImageFilesystem *fs, IsoFileSource *parent,
          * It BREAKS POSIX SEMANTICS, but its suitable for our needs
          */
 
+#ifndef Libisofs_ino_from_lbA
 #define Libisofs_patch_ticket_144 yes
+#endif
 #ifdef Libisofs_patch_ticket_144
-        /* Trying to ensure unique inode numbers */
-        {
-            static ino_t iso_global_inode= 0;
 
-            if(iso_global_inode <= 0)
-                iso_global_inode= 1;
-            atts.st_ino = iso_global_inode++;
-        }
+        atts.st_ino = fs_give_ino_number(fs, 0);
+
 #else
         /* Ticket 144: This produces duplicate numbers with empty files.
         */
@@ -1428,6 +1538,8 @@ int iso_file_source_new_ifs(IsoImageFilesystem *fs, IsoFileSource *parent,
         }
     }
 
+#endif /* ! Libisofs_new_fs_image_inO */
+
     /*
      * if we haven't RR extensions, or a needed TF time stamp is not present,
      * we use plain iso recording time
@@ -1487,6 +1599,7 @@ int iso_file_source_new_ifs(IsoImageFilesystem *fs, IsoFileSource *parent,
     }
     ifsdata->info = atts;
     ifsdata->name = name;
+    ifsdata->aa_string = aa_string;
 
     /* save extents */
     ifsdata->sections = realloc(ifsdata->sections,
@@ -1830,17 +1943,6 @@ int read_root_susp_entries(_ImageFsData *data, uint32_t block)
             continue;
 
         if (SUSP_SIG(sue, 'E', 'R')) {
-
-            if (data->rr_version) {
-                ret = iso_msg_submit(data->msgid, ISO_SUSP_MULTIPLE_ER, 0,
-                    "More than one ER has found. This is not supported. "
-                    "It will be ignored, but can cause problems. "
-                    "Please notify us about this.");
-                if (ret < 0) {
-                    break;
-                }
-            }
-
             /*
              * it seems that Rock Ridge can be identified with any
              * of the following
@@ -1860,11 +1962,28 @@ int read_root_susp_entries(_ImageFsData *data, uint32_t block)
                 iso_msg_debug(data->msgid,
                               "Suitable Rock Ridge ER found. Version 1.12.");
                 data->rr_version = RR_EXT_112;
+
+            } else if (sue->data.ER.len_id[0] == 9 &&
+                  (strncmp((char*)sue->data.ER.ext_id, "AAIP_0002", 9) == 0 ||
+                   strncmp((char*)sue->data.ER.ext_id, "AAIP_0100", 9) == 0)) {
+
+                /* Tolerate AAIP ER even if not supported */
+                iso_msg_debug(data->msgid,
+                              "Suitable AAIP ER found.");
+
+                if (((char*)sue->data.ER.ext_id)[6] == '1')
+                    data->aaip_version = 100;
+                else
+                    data->aaip_version = 2;
+                if (!data->aaip_load)
+                    iso_msg_submit(data->msgid, ISO_AAIP_IGNORED, 0,
+           "Identifier for extension AAIP found, but loading is not enabled.");
+
             } else {
                 ret = iso_msg_submit(data->msgid, ISO_SUSP_MULTIPLE_ER, 0,
-                    "Not Rock Ridge ER found.\n"
-                    "That will be ignored, but can cause problems in "
-                    "image reading. Please notify us about this");
+                    "Unknown Extension Signature found in ER.\n"
+                    "It will be ignored, but can cause problems in "
+                    "image reading. Please notify us about this.");
                 if (ret < 0) {
                     break;
                 }
@@ -2026,6 +2145,8 @@ int iso_image_filesystem_new(IsoDataSource *src, struct iso_read_opts *opts,
     data->file_mode = opts->file_mode & ~S_IFMT;
     data->dir_mode = opts->dir_mode & ~S_IFMT;
     data->msgid = msgid;
+    data->aaip_load = !opts->noaaip;
+    data->aaip_version = -1;
 
     /* ??? ts Nov 25 2008 :
        Shouldn't this go to library initialization or even to app ?
@@ -2223,6 +2344,46 @@ int iso_image_filesystem_new(IsoDataSource *src, struct iso_read_opts *opts,
     return ret;
 }
 
+
+/* Take over aa_string from file source to node or discard it after making
+   the necessary change in node->mode group permissions.
+   node->mode must already be set.
+*/
+static
+int src_aa_to_node(IsoFileSource *src, IsoNode *node, int flag)
+{
+    int ret;
+    unsigned char *aa_string;
+    ImageFileSourceData *data;
+    _ImageFsData *fsdata;
+    char *a_text = NULL, *d_text = NULL;
+
+    data = (ImageFileSourceData*)src->data;
+    fsdata = data->fs->data;
+
+    /* Obtain ownership of eventual AA string */
+    ret = iso_file_source_get_aa_string(src, &aa_string, 1);
+    if (ret != 1 || aa_string == NULL)
+        return 1;
+    if (fsdata->aaip_load == 1) {
+        /* Attach aa_string to node */
+        ret = iso_node_add_xinfo(node, aaip_xinfo_func, aa_string);
+        if (ret < 0)
+            return ret;
+    } else {
+        /* Look for ACL and perform S_IRWXG mapping */
+        iso_aa_get_acl_text(aa_string, node->mode, &a_text, &d_text, 16);
+        if (a_text != NULL)
+            aaip_cleanout_st_mode(a_text, &(node->mode), 4 | 16);
+        /* Dispose ACL a_text and d_text */
+        iso_aa_get_acl_text(aa_string, node->mode, &a_text, &d_text, 1 << 15);
+        /* Dispose aa_string */
+        aaip_xinfo_func(aa_string, 1);
+    }
+    return 1;
+}
+
+
 static
 int image_builder_create_node(IsoNodeBuilder *builder, IsoImage *image,
                               IsoFileSource *src, IsoNode **node)
@@ -2402,6 +2563,15 @@ int image_builder_create_node(IsoNodeBuilder *builder, IsoImage *image,
     new->parent = NULL;
     new->next = NULL;
 
+    ret = src_aa_to_node(src, new, 0);
+    if (ret < 0) {
+        /* todo: stuff any possible memory leak here */
+        if (name != NULL)
+            free(name);
+        free(new);
+        return ret;
+    }
+
     *node = new;
     return ISO_SUCCESS;
 }
@@ -2456,7 +2626,17 @@ int create_boot_img_filesrc(IsoImageFilesystem *fs, IsoFileSource **src)
 
     memset(&atts, 0, sizeof(struct stat));
     atts.st_mode = S_IFREG;
+
+#ifdef Libisofs_new_fs_image_inO
+
+    atts.st_ino = fs_give_ino_number(fs, 0);
+
+#else /* Libisofs_new_fs_image_inO */
+
     atts.st_ino = fsdata->imgblock; /* not the best solution, but... */
+
+#endif /* ! Libisofs_new_fs_image_inO */
+
     atts.st_nlink = 1;
 
     /*
@@ -2576,6 +2756,12 @@ int iso_image_import(IsoImage *image, IsoDataSource *src,
         image->root->node.atime = info.st_atime;
         image->root->node.mtime = info.st_mtime;
         image->root->node.ctime = info.st_ctime;
+
+        /* This might fail in iso_node_add_xinfo() */
+        ret = src_aa_to_node(newroot, &(image->root->node), 0);
+        if (ret < 0)
+            goto import_revert;
+
     }
 
     /* if old image has el-torito, add a new catalog */
@@ -2777,6 +2963,8 @@ int iso_read_opts_new(IsoReadOpts **opts, int profile)
 
     ropts->file_mode = 0444;
     ropts->dir_mode = 0555;
+    ropts->noaaip= 1;
+
     *opts = ropts;
     return ISO_SUCCESS;
 }
@@ -2827,6 +3015,15 @@ int iso_read_opts_set_no_iso1999(IsoReadOpts *opts, int noiso1999)
     return ISO_SUCCESS;
 }
 
+int iso_read_opts_set_no_aaip(IsoReadOpts *opts, int noaaip)
+{
+    if (opts == NULL) {
+        return ISO_NULL_POINTER;
+    }
+    opts->noaaip = noaaip ? 1 : 0;
+    return ISO_SUCCESS;
+}
+
 int iso_read_opts_set_preferjoliet(IsoReadOpts *opts, int preferjoliet)
 {
     if (opts == NULL) {
