@@ -1,6 +1,7 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
  * Copyright (c) 2007 Mario Danic
+ * Copyright (c) 2009 Thomas Schmitt
  * 
  * This file is part of the libisofs project; you can redistribute it and/or 
  * modify it under the terms of the GNU General Public License version 2 as 
@@ -13,9 +14,15 @@
 #include "writer.h"
 #include "messages.h"
 #include "image.h"
+#include "aaip_0_2.h"
 
 #include <string.h>
 
+
+static
+int susp_add_ES(Ecma119Image *t, struct susp_info *susp, int to_ce, int seqno);
+
+
 static
 int susp_append(Ecma119Image *t, struct susp_info *susp, uint8_t *data)
 {
@@ -374,6 +381,38 @@ int rrip_SL_append_comp(size_t *n, uint8_t ***comps, char *s, int size, char fl)
     return ISO_SUCCESS;
 }
 
+
+#ifdef Libisofs_with_rrip_rR
+
+/* ts A90307 */
+/**
+ * Add to the given tree node a RR System Use Entry. This is an obsolete 
+ * entry from before RRIP-1.10. Nevertheless mkisofs produces it and there
+ * is the suspicion that Solaris takes it as indication for Rock Ridge.
+ *
+ * I once saw a copy of a RRIP spec which mentioned RR. Here i just use
+ * the same constant 5 bytes as produced by mkisofs.
+ */
+static
+int rrip_add_RR(Ecma119Image *t, Ecma119Node *n, struct susp_info *susp)
+{
+    uint8_t *RR;
+    RR = malloc(5);
+    if (RR == NULL) {
+        return ISO_OUT_OF_MEM;
+    }
+
+    RR[0] = 'R';
+    RR[1] = 'R';
+    RR[2] = 5;
+    RR[3] = 1;
+    RR[4] = 0201;
+    return susp_append(t, susp, RR);
+}
+
+#endif /* Libisofs_with_rrip_rR */
+
+
 /**
  * Add a SL System Use Entry to the given tree node. This is used to store 
  * the content of a symbolic link, and is mandatory if the tree node
@@ -460,6 +499,69 @@ int rrip_add_SL(Ecma119Image *t, struct susp_info *susp, uint8_t **comp,
     return ret;
 }
 
+
+/*
+   @param flag bit0= only account sizes in sua_free resp. ce_len
+                     parameters susp and data may be NULL in this case
+*/
+static
+int aaip_add_AA(Ecma119Image *t, struct susp_info *susp,
+                uint8_t **data, size_t num_data,
+                size_t *sua_free, size_t *ce_len, int flag)
+{
+    int ret, done = 0, len, es_extra = 0;
+    uint8_t *aapt, *cpt;
+
+    if (!t->aaip_susp_1_10)
+        es_extra = 5;
+    if (*sua_free < num_data + es_extra || *ce_len > 0) {
+        *ce_len += num_data + es_extra;
+    } else {
+        *sua_free -= num_data + es_extra;
+    }
+    if (flag & 1)
+        return ISO_SUCCESS;
+
+    /* If AAIP enabled and announced by ER : Write ES field to announce AAIP */
+    if (t->aaip && !t->aaip_susp_1_10) {
+        ret = susp_add_ES(t, susp, (*ce_len > 0), 1);
+        if (ret < 0)
+            return ret;
+    }
+    aapt = *data;
+    if (!(aapt[4] & 1)) {
+        /* Single field can be handed over directly */
+        if (*ce_len > 0) {
+            ret = susp_append_ce(t, susp, aapt);
+        } else {
+            ret = susp_append(t, susp, aapt);
+        }
+        *data = NULL;
+        return ISO_SUCCESS;
+    }
+
+    /* Multiple fields have to be handed over as single field copies */
+    for (aapt = *data; !done; aapt += aapt[2]) {
+        done = !(aapt[4] & 1);
+        len = aapt[2];
+        cpt = calloc(aapt[2], 1);
+        if (cpt == NULL)
+            return ISO_OUT_OF_MEM;
+        memcpy(cpt, aapt, len);
+        if (*ce_len > 0) {
+            ret = susp_append_ce(t, susp, cpt);
+        } else {
+            ret = susp_append(t, susp, cpt);
+        }
+        if (ret == -1)
+            return ret;
+    }
+    free(*data);
+    *data = NULL;
+    return ISO_SUCCESS;
+}
+
+
 /**
  * Add a SUSP "ER" System Use Entry to identify the Rock Ridge specification.
  * 
@@ -483,6 +585,7 @@ int rrip_add_ER(Ecma119Image *t, struct susp_info *susp)
          decrease by 1."
         So "IEEE_P1282" would be the new form, "RRIP_1991A" is the old form.
         */
+
         ER = malloc(182);
         if (ER == NULL) {
             return ISO_OUT_OF_MEM;
@@ -540,6 +643,52 @@ int rrip_add_ER(Ecma119Image *t, struct susp_info *susp)
     return susp_append_ce(t, susp, ER);
 }
 
+
+static
+int aaip_add_ER(Ecma119Image *t, struct susp_info *susp, int flag)
+{
+    unsigned char *AA;
+
+    AA = malloc(160);
+    if (AA == NULL) {
+        return ISO_OUT_OF_MEM;
+    }
+    
+    AA[0] = 'E';
+    AA[1] = 'R';
+    AA[2] = 160;
+    AA[3] = 1;
+    AA[4] = 9;
+    AA[5] = 81;
+    AA[6] = 62;
+    AA[7] = 1;
+
+#define Libisofs_aaip_1_0 yes
+#ifdef Libisofs_aaip_1_0
+
+    memcpy(AA + 8, "AAIP_0100", 9);
+    memcpy(AA + 17,
+           "AA PROVIDES VIA AAIP 1.0 SUPPORT FOR ARBITRARY FILE ATTRIBUTES"
+           " IN ISO 9660 IMAGES", 81);
+
+#else /* Libisofs_aaip_1_0 */
+
+    memcpy(AA + 8, "AAIP_0002", 9);
+    memcpy(AA + 17,
+           "AA PROVIDES VIA AAIP 0.2 SUPPORT FOR ARBITRARY FILE ATTRIBUTES"
+           " IN ISO 9660 IMAGES", 81);
+
+#endif /* ! Libisofs_aaip_1_0 */
+
+    memcpy(AA + 98,
+           "PLEASE CONTACT THE LIBBURNIA PROJECT VIA LIBBURNIA-PROJECT.ORG",
+           62);
+
+    /** This always goes to continuation area */
+    return susp_append_ce(t, susp, AA);
+}
+
+
 /**
  * Add a CE System Use Entry to the given tree node. A "CE" is used to add
  * a continuation area, where additional System Use Entry can be written.
@@ -587,6 +736,226 @@ int susp_add_SP(Ecma119Image *t, struct susp_info *susp)
     return susp_append(t, susp, SP);
 }
 
+
+/**
+ * SUSP 1.12: [...] shall specify as an 8-bit number the Extension
+ * Sequence Number of the extension specification utilized in the entries
+ * immediately following this System Use Entry. The Extension Sequence
+ * Numbers of multiple extension specifications on a volume shall correspond to
+ * the order in which their "ER" System Use Entries are recorded [...]
+ */
+static
+int susp_add_ES(Ecma119Image *t, struct susp_info *susp, int to_ce, int seqno)
+{
+    unsigned char *ES = malloc(5);
+
+    if (ES == NULL) {
+        return ISO_OUT_OF_MEM;
+    }
+    ES[0] = 'E';
+    ES[1] = 'S';
+    ES[2] = (unsigned char) 5;
+    ES[3] = (unsigned char) 1;
+    ES[4] = (unsigned char) seqno;
+    if (to_ce) {
+        return susp_append_ce(t, susp, ES);
+    } else {
+        return susp_append(t, susp, ES);
+    }
+}
+
+
+int aaip_xinfo_func(void *data, int flag)
+{
+    if (flag & 1) {
+        free(data);
+    }
+    return 1;
+}
+
+
+/**
+ * Compute SUA lentgth and eventual Continuation Area length of field NM and
+ * eventually fields SL and AA. Because CA usage makes necessary the use of
+ * a CE entry of 28 bytes in SUA, this computation fails if not the 28 bytes
+ * are taken into account at start. In this case the caller should retry with
+ * bit0 set. 
+ * 
+ * @param flag      bit0= assume CA usage (else return 0 on SUA overflow)
+ * @return          1= ok, computation of *su_size and *ce is valid
+ *                  0= not ok, CA usage is necessary but bit0 was not set
+ *                     (*su_size and *ce stay unaltered in this case)
+ *                 <0= error:
+ *                 -1= not enough SUA space for 28 bytes of CE entry
+ */
+static
+int susp_calc_nm_sl_aa(Ecma119Image *t, Ecma119Node *n, size_t space,
+                       size_t *su_size, size_t *ce, int flag)
+{
+    char *name;
+    size_t namelen, su_mem, ce_mem;
+    void *xipt;
+    size_t num_aapt = 0, sua_free = 0;
+    int ret;
+
+    su_mem = *su_size;
+    ce_mem = *ce;
+    if (*ce > 0 && !(flag & 1))
+        goto unannounced_ca;
+
+    name = get_rr_fname(t, n->node->name);
+    namelen = strlen(name);
+    free(name);
+
+    if (flag & 1) {
+       /* Account for 28 bytes of CE field */
+       if (*su_size + 28 > space)
+           return -1;
+       *su_size += 28;
+    }
+
+    /* NM entry */
+    if (*su_size + 5 + namelen <= space) {
+        /* ok, it fits in System Use Area */
+        *su_size += 5 + namelen;
+    } else {
+        /* the NM will be divided in a CA */
+        if (!(flag & 1))
+            goto unannounced_ca;
+        namelen = namelen - (space - *su_size - 5);
+        *ce = 5 + namelen;
+        *su_size = space;
+    }
+    if (n->type == ECMA119_SYMLINK) {
+        /* 
+         * for symlinks, we also need to write the SL
+         */
+        char *dest, *cur, *prev;
+        size_t sl_len = 5;
+        int cew = (*ce != 0); /* are we writing to CA ? */
+
+        dest = get_rr_fname(t, ((IsoSymlink*)n->node)->dest);
+        prev = dest;
+        cur = strchr(prev, '/');
+        while (1) {
+            size_t clen;
+            if (cur) {
+                clen = cur - prev;
+            } else {
+                /* last component */
+                clen = strlen(prev);
+            }
+
+            if (clen == 1 && prev[0] == '.') {
+                clen = 0;
+            } else if (clen == 2 && prev[0] == '.' && prev[1] == '.') {
+                clen = 0;
+            }
+
+            /* flags and len for each component record (RRIP, 4.1.3.1) */
+            clen += 2;
+
+            if (!cew) {
+                /* we are still writing to the SUA */
+                if (*su_size + sl_len + clen > space) {
+                    /* 
+                     * ok, we need a Continuation Area anyway
+                     * TODO this can be handled better, but for now SL
+                     * will be completelly moved into the CA
+                     */
+                    if (!(flag & 1))
+                        goto unannounced_ca;
+                    cew = 1;
+                } else {
+                    sl_len += clen;
+                }
+            }
+            if (cew) {
+                if (sl_len + clen > 255) {
+                    /* we need an additional SL entry */
+                    if (clen > 250) {
+                        /* 
+                         * case 1, component too large to fit in a 
+                         * single SL entry. Thus, the component need
+                         * to be divided anyway.
+                         * Note than clen can be up to 255 + 2 = 257.
+                         * 
+                         * First, we check how many bytes fit in current
+                         * SL field 
+                         */
+                        int fit = 255 - sl_len - 2;
+                        if (clen - 250 <= fit) {
+                            /* 
+                             * the component can be divided between this
+                             * and another SL entry
+                             */
+                            *ce += 255; /* this SL, full */
+                            sl_len = 5 + (clen - fit);
+                        } else {
+                            /*
+                             * the component will need a 2rd SL entry in
+                             * any case, so we prefer to don't write 
+                             * anything in this SL
+                             */
+                            *ce += sl_len + 255;
+                            sl_len = 5 + (clen - 250) + 2;
+                        }
+                    } else {
+                        /* case 2, create a new SL entry */
+                        *ce += sl_len;
+                        sl_len = 5 + clen;
+                    }
+                } else {
+                    sl_len += clen;
+                }
+            }
+
+            if (!cur || cur[1] == '\0') {
+                /* cur[1] can be \0 if dest ends with '/' */
+                break;
+            }
+            prev = cur + 1;
+            cur = strchr(prev, '/');
+        }
+
+        free(dest);
+
+        /* and finally write the pending SL field */
+        if (!cew) {
+            /* the whole SL fits into the SUA */
+            *su_size += sl_len;
+        } else {
+            *ce += sl_len;
+        }
+
+    }
+
+    /* obtain num_aapt from node */
+    xipt = NULL;
+    num_aapt = 0;
+    if (t->aaip) {
+        ret = iso_node_get_xinfo(n->node, aaip_xinfo_func, &xipt);
+        if (ret == 1) {
+           num_aapt = aaip_count_bytes((unsigned char *) xipt, 0);
+        }
+    }
+    /* let the expert decide where to add num_aapt */
+    if (num_aapt > 0) {
+        sua_free = space - *su_size;
+        aaip_add_AA(t, NULL, NULL, num_aapt, &sua_free, ce, 1);
+        *su_size = space - sua_free;
+        if (*ce > 0 && !(flag & 1))
+            goto unannounced_ca;
+    }
+    return 1;
+
+unannounced_ca:;
+    *su_size = su_mem;
+    *ce = ce_mem;
+    return 0;
+}
+
+
 /**
  * Compute the length needed for write all RR and SUSP entries for a given
  * node.
@@ -605,6 +974,7 @@ size_t rrip_calc_len(Ecma119Image *t, Ecma119Node *n, int type, size_t space,
                      size_t *ce)
 {
     size_t su_size;
+    int ret;
 
     /* space min is 255 - 33 - 37 = 185
      * At the same time, it is always an odd number, but we need to pad it
@@ -614,11 +984,22 @@ size_t rrip_calc_len(Ecma119Image *t, Ecma119Node *n, int type, size_t space,
     space--;
     *ce = 0;
 
+    su_size = 0;
+
+    /* If AAIP enabled and announced by ER : account for 5 bytes of ES */;
+    if (t->aaip && !t->aaip_susp_1_10)
+        su_size += 5;
+
+#ifdef Libisofs_with_rrip_rR
+    /* obsolete RR field (once in AAIP-1.09) */
+    su_size += 5;
+#endif
+
     /* PX and TF, we are sure they always fit in SUA */
     if (!t->rrip_version_1_10) {
-        su_size = 44 + 26;
+        su_size += 44 + 26;
     } else {
-        su_size = 36 + 26;
+        su_size += 36 + 26;
     }
 
     if (n->type == ECMA119_DIR) {
@@ -643,133 +1024,18 @@ size_t rrip_calc_len(Ecma119Image *t, Ecma119Node *n, int type, size_t space,
     }
 
     if (type == 0) {
-        char *name = get_rr_fname(t, n->node->name);
-        size_t namelen = strlen(name);
-        free(name);
 
-        /* NM entry */
-        if (su_size + 5 + namelen <= space) {
-            /* ok, it fits in System Use Area */
-            su_size += 5 + namelen;
-        } else {
-            /* the NM will be divided in a CE */
-            namelen = namelen - (space - su_size - 5 - 28);
-            *ce = 5 + namelen;
-            su_size = space;
-        }
-        if (n->type == ECMA119_SYMLINK) {
-            /* 
-             * for symlinks, we also need to write the SL
-             */
-            char *dest, *cur, *prev;
-            size_t sl_len = 5;
-            int cew = (*ce != 0); /* are we writing to CE? */
+        /* Try without CE */
+        ret = susp_calc_nm_sl_aa(t, n, space, &su_size, ce, 0);
+        if (ret == 0) /* Retry with CE */
+            susp_calc_nm_sl_aa(t, n, space, &su_size, ce, 1);
 
-            dest = get_rr_fname(t, ((IsoSymlink*)n->node)->dest);
-            prev = dest;
-            cur = strchr(prev, '/');
-            while (1) {
-                size_t clen;
-                if (cur) {
-                    clen = cur - prev;
-                } else {
-                    /* last component */
-                    clen = strlen(prev);
-                }
-
-                if (clen == 1 && prev[0] == '.') {
-                    clen = 0;
-                } else if (clen == 2 && prev[0] == '.' && prev[1] == '.') {
-                    clen = 0;
-                }
-
-                /* flags and len for each component record (RRIP, 4.1.3.1) */
-                clen += 2;
-
-                if (!cew) {
-                    /* we are still writing to the SUA */
-                    if (su_size + sl_len + clen > space) {
-                        /* 
-                         * ok, we need a Continuation Area anyway
-                         * TODO this can be handled better, but for now SL
-                         * will be completelly moved into the CA
-                         */
-                        if (su_size + 28 <= space) {
-                            /* the CE entry fills without reducing NM */
-                            su_size += 28;
-                        } else {
-                            /* we need to reduce NM */
-                            *ce = (28 - (space - su_size)) + 5;
-                            su_size = space;
-                        }
-                        cew = 1;
-                    } else {
-                        sl_len += clen;
-                    }
-                }
-                if (cew) {
-                    if (sl_len + clen > 255) {
-                        /* we need an additional SL entry */
-                        if (clen > 250) {
-                            /* 
-                             * case 1, component too large to fit in a 
-                             * single SL entry. Thus, the component need
-                             * to be divided anyway.
-                             * Note than clen can be up to 255 + 2 = 257.
-                             * 
-                             * First, we check how many bytes fit in current
-                             * SL field 
-                             */
-                            int fit = 255 - sl_len - 2;
-                            if (clen - 250 <= fit) {
-                                /* 
-                                 * the component can be divided between this
-                                 * and another SL entry
-                                 */
-                                *ce += 255; /* this SL, full */
-                                sl_len = 5 + (clen - fit);
-                            } else {
-                                /*
-                                 * the component will need a 2rd SL entry in
-                                 * any case, so we prefer to don't write 
-                                 * anything in this SL
-                                 */
-                                *ce += sl_len + 255;
-                                sl_len = 5 + (clen - 250) + 2;
-                            }
-                        } else {
-                            /* case 2, create a new SL entry */
-                            *ce += sl_len;
-                            sl_len = 5 + clen;
-                        }
-                    } else {
-                        sl_len += clen;
-                    }
-                }
-
-                if (!cur || cur[1] == '\0') {
-                    /* cur[1] can be \0 if dest ends with '/' */
-                    break;
-                }
-                prev = cur + 1;
-                cur = strchr(prev, '/');
-            }
-
-            free(dest);
-
-            /* and finally write the pending SL field */
-            if (!cew) {
-                /* the whole SL fits into the SUA */
-                su_size += sl_len;
-            } else {
-                *ce += sl_len;
-            }
-
-        }
     } else {
 
         /* "." or ".." entry */
+
         su_size += 5; /* NM field */
+
         if (type == 1 && n->parent == NULL) {
             /* 
              * "." for root directory 
@@ -777,7 +1043,23 @@ size_t rrip_calc_len(Ecma119Image *t, Ecma119Node *n, int type, size_t space,
              * ER needs a Continuation Area, thus we also need a CE entry
              */
             su_size += 7 + 28; /* SP + CE */
-            *ce = 182; /* ER */
+
+#ifdef Libisofs_rrip_1_10_er_bugfiX
+
+            /* ER of RRIP */
+            if (t->rrip_version_1_10) {
+                *ce = 237;
+            } else {
+                *ce = 182;
+            }
+
+#else
+            *ce = 182;
+#endif
+
+            if (t->aaip) {
+                *ce += 160; /* ER of AAIP */
+            }
         }
     }
 
@@ -808,6 +1090,37 @@ void susp_info_free(struct susp_info* susp)
     free(susp->ce_susp_fields);
 }
 
+
+/* @param flag bit0= Do not add data but only count sua_free and ce_len
+*/
+static
+int add_aa_string(Ecma119Image *t, Ecma119Node *n, struct susp_info *info,
+                  size_t *sua_free, size_t *ce_len, int flag)
+{
+    int ret;
+    uint8_t *aapt;
+    void *xipt;
+    size_t num_aapt= 0;
+
+    ret = iso_node_get_xinfo(n->node, aaip_xinfo_func, &xipt);
+    if (ret == 1) {
+        num_aapt = aaip_count_bytes((unsigned char *) xipt, 0);
+        if (num_aapt > 0) {
+            aapt = malloc(num_aapt);
+            if (aapt == NULL)
+                return ISO_OUT_OF_MEM;
+            memcpy(aapt, xipt, num_aapt);
+            ret = aaip_add_AA(t, info, &aapt, num_aapt, sua_free, ce_len,
+                              flag & 1);
+            if (ret < 0) 
+                return ret;
+            /* aapt is NULL now and the memory is owned by t */
+        }
+    }
+    return 1;
+}
+
+
 /**
  * Fill a struct susp_info with the RR/SUSP entries needed for a given
  * node.
@@ -833,6 +1146,15 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
     Ecma119Node *node;
     char *name = NULL;
     char *dest = NULL;
+    size_t aaip_er_len= 0;
+
+#ifdef Libisofs_rrip_1_10_er_bugfiX
+    size_t rrip_er_len= 182;
+#endif
+
+    size_t su_size_pd, ce_len_pd; /* predicted sizes of SUA and CA */
+    int ce_is_predicted = 0;
+    size_t aaip_sua_free= 0, aaip_len= 0;
 
     if (t == NULL || n == NULL || info == NULL) {
         return ISO_NULL_POINTER;
@@ -866,6 +1188,21 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
         }
     }
 
+    /* If AAIP enabled and announced by ER : Announce RRIP by ES */
+    if (t->aaip && !t->aaip_susp_1_10) {
+        ret = susp_add_ES(t, info, 0, 0);
+        if (ret < 0)
+            goto add_susp_cleanup;
+    }
+
+#ifdef Libisofs_with_rrip_rR
+    /* ts A90307 */
+    ret = rrip_add_RR(t, node, info);
+    if (ret < 0) {
+        goto add_susp_cleanup;
+    }
+#endif /* Libisofs_with_rrip_rR */
+
     /* PX and TF, we are sure they always fit in SUA */
     ret = rrip_add_PX(t, node, info);
     if (ret < 0) {
@@ -927,6 +1264,16 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
 
         sua_free = space - info->suf_len;
 
+        /* Try whether NM, SL, AA will fit into SUA */
+        su_size_pd = info->suf_len;
+        ce_len_pd = ce_len;
+        ret = susp_calc_nm_sl_aa(t, n, space, &su_size_pd, &ce_len_pd, 0);
+        if (ret == 0) { /* Have to use CA. 28 bytes of CE are necessary */
+            susp_calc_nm_sl_aa(t, n, space, &su_size_pd, &ce_len_pd, 1);
+            sua_free -= 28;
+            ce_is_predicted = 1;
+        }
+
         /* NM entry */
         if (5 + namelen <= sua_free) {
             /* ok, it fits in System Use Area */
@@ -935,7 +1282,7 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
         } else {
             /* the NM will be divided in a CE */
             nm_type = 1;
-            namelen = namelen - (sua_free - 5 - 28);
+            namelen = namelen - (sua_free - 5);
             ce_len = 5 + namelen;
             sua_free = 0;
         }
@@ -983,17 +1330,10 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
                          * TODO this can be handled better, but for now SL
                          * will be completelly moved into the CA
                          */
-                        if (28 <= sua_free) {
-                            /* the CE entry fills without reducing NM */
-                            sua_free -= 28;
-                            cew = 1;
-                        } else {
-                            /* we need to reduce NM */
-                            nm_type = 1;
-                            ce_len = (28 - sua_free) + 5;
-                            sua_free = 0;
-                            cew = 1;
-                        }
+
+                        /* sua_free, ce_len, nm_type already account for CE */
+                        cew = 1;
+
                     } else {
                         /* add the component */
                         ret = rrip_SL_append_comp(&n_comp, &comps, prev, 
@@ -1110,16 +1450,17 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
              * Write the NM part that fits in SUA...  Note that CE
              * entry and NM in the continuation area is added below 
              */
-            namelen = space - info->suf_len - 28 - 5;
+
+            namelen = space - info->suf_len - 28 * (!!ce_is_predicted) - 5;
             ret = rrip_add_NM(t, info, name, namelen, 1, 0);
             if (ret < 0) {
                 goto add_susp_cleanup;
             }
         }
 
-        if (ce_len > 0) {
+        if (ce_is_predicted) {
             /* Add the CE entry */
-            ret = susp_add_CE(t, ce_len, info);
+            ret = susp_add_CE(t, ce_len_pd, info);
             if (ret < 0) {
                 goto add_susp_cleanup;
             }
@@ -1152,6 +1493,17 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
             }
         }
 
+        /* Obtain AA field string from node
+           and write it to directory entry or CE area.
+        */
+        ret = ISO_SUCCESS;
+
+        if (t->aaip) {
+            ret = add_aa_string(t, n, info, &sua_free, &ce_len, 0);
+            if (ret < 0)
+                goto add_susp_cleanup;
+        }
+
     } else {
 
         /* "." or ".." entry */
@@ -1161,6 +1513,7 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
         if (ret < 0) {
             goto add_susp_cleanup;
         }
+
         if (type == 1 && n->parent == NULL) {
             /* 
              * "." for root directory 
@@ -1168,7 +1521,45 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
              * ER needs a Continuation Area, thus we also need a CE entry.
              * Note that SP entry was already added above
              */
-            ret = susp_add_CE(t, 182, info); /* 182 is ER length */
+
+#ifdef Libisofs_rrip_1_10_er_bugfiX
+
+            if (t->rrip_version_1_10) {
+                rrip_er_len = 237;
+            } else {
+                rrip_er_len = 182;
+            }
+            if (t->aaip && !t->aaip_susp_1_10) {
+                aaip_er_len = 160;
+            }
+
+            /* Compute length of AAIP string of root node */
+            aaip_sua_free= 0;
+            ret = add_aa_string(t, n, info, &aaip_sua_free, &aaip_len, 1);
+            if (ret < 0)
+                goto add_susp_cleanup;
+
+            /* Allocate the necessary CE space */
+            ret = susp_add_CE(t, rrip_er_len + aaip_er_len + aaip_len, info);
+
+#else /* Libisofs_rrip_1_10_er_bugfiX */
+
+            if (t->aaip && !t->aaip_susp_1_10) {
+                aaip_er_len = 160;
+            }
+
+            /* Compute length of AAIP string of root node */
+            aaip_sua_free= 0;
+            ret = add_aa_string(t, n, info, &aaip_sua_free, &aaip_len, 1);
+            if (ret < 0)
+                goto add_susp_cleanup;
+
+            /* Allocate the necessary CE space */
+            ret = susp_add_CE(t, 182 + aaip_er_len + aaip_len, info);
+                                                    /* 182 is RRIP-ER length */
+
+#endif /* ! Libisofs_rrip_1_10_er_bugfiX */
+
             if (ret < 0) {
                 goto add_susp_cleanup;
             }
@@ -1176,9 +1567,22 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
             if (ret < 0) {
                 goto add_susp_cleanup;
             }
+            if (t->aaip && !t->aaip_susp_1_10) {
+                ret = aaip_add_ER(t, info, 0);
+                if (ret < 0) {
+                    goto add_susp_cleanup;
+                }
+            }
+            /* Write AAIP string of root node */
+            aaip_sua_free= aaip_len= 0;
+            ret = add_aa_string(t, n, info, &aaip_sua_free, &aaip_len, 0);
+            if (ret < 0)
+                goto add_susp_cleanup;
+
         }
     }
 
+
     /*
      * The System Use field inside the directory record must be padded if
      * it is an odd number (ECMA-119, 9.1.13)
