@@ -1,6 +1,7 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
  * Copyright (c) 2007 Mario Danic
+ * Copyright (c) 2009 Thomas Schmitt
  * 
  * This file is part of the libisofs project; you can redistribute it and/or 
  * modify it under the terms of the GNU General Public License version 2 as 
@@ -28,6 +29,7 @@
 
 #include "ecma119.h"
 
+
 #define SUSP_SIG(entry, a, b) ((entry->sig[0] == a) && (entry->sig[1] == b))
 
 /**
@@ -115,6 +117,14 @@ struct rr_SL {
     uint8_t comps[1];
 };
 
+
+/** Arbitrary Attribute (AAIP, see doc/susp_aaip_1_0.txt) */
+struct rr_AA {
+    uint8_t flags[1];
+    uint8_t comps[1];
+};
+
+
 /**
  * Struct for a SUSP System User Entry (SUSP, 4.1)
  */
@@ -133,6 +143,8 @@ struct susp_sys_user_entry
         struct rr_NM NM;
         struct rr_CL CL;
         struct rr_SL SL;
+        struct rr_AA AA;
+
     } data; /* 5 to 4+len_sue */
 };
 
@@ -264,4 +276,26 @@ int read_rr_SL(struct susp_sys_user_entry *sl, char **dest, int *cont);
  */
 int read_rr_PN(struct susp_sys_user_entry *pn, struct stat *st);
 
+
+/**
+ * Collects the AA field string from single AA fields.
+ * (see doc/susp_aaip_0_2.txt)
+ * @param aa_string   Storage location of the emerging string.
+ *                    Begin with *aa_string == NULL, or own malloc() storage.
+ * @param aa_size     Current allocated size of aa_string.
+ *                    Begin with *aa_size == 0, or own storage size.
+ * @param aa_len      Current occupied size of aa_string.
+ *                    Begin with *aa_len == 0
+ * @param prev_field  Returns the index of start of the previous field
+ *                    in the string.
+ * @param is_done     The current completion state of the AA field string.
+ *                    Fields will be ignored as soon as it is 1.
+ *                    Begin with *is_done == 0
+ * @param flag        Unused yet. Submit 0.
+ */
+int read_aaip_AA(struct susp_sys_user_entry *sue,
+                 unsigned char **aa_string, size_t *aa_size, size_t *aa_len,
+                 size_t *prev_field, int *is_done, int flag);
+
+
 #endif /* LIBISO_ROCKRIDGE_H */
