@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
+ * Copyright (c) 2009 Thomas Schmitt
  * 
  * This file is part of the libisofs project; you can redistribute it and/or 
  * modify it under the terms of the GNU General Public License version 2 as 
@@ -7,8 +8,8 @@
  */
 
 /*
- * This file contains functions related to the reading of SUSP and
- * Rock Ridge extensions on an ECMA-119 image.
+ * This file contains functions related to the reading of SUSP, 
+ * Rock Ridge and AAIP extensions on an ECMA-119 image.
  */
 
 #include "libisofs.h"
@@ -402,6 +403,8 @@ int read_rr_SL(struct susp_sys_user_entry *sl, char **dest, int *cont)
  */
 int read_rr_PN(struct susp_sys_user_entry *pn, struct stat *st)
 {
+    int high_shift= 0;
+
     if (pn == NULL || pn == NULL) {
         return ISO_NULL_POINTER;
     }
@@ -413,7 +416,67 @@ int read_rr_PN(struct susp_sys_user_entry *pn, struct stat *st)
         return ISO_WRONG_RR;
     }
 
+    /* (dev_t << 32) causes compiler warnings on FreeBSD
+        because sizeof(dev_t) is 4.
+    */
+    st->st_rdev = (dev_t)iso_read_bb(pn->data.PN.low, 4, NULL);
+    if (sizeof(st->st_rdev) > 4) {
+        high_shift = 32;
+        st->st_rdev |= (dev_t)((dev_t)iso_read_bb(pn->data.PN.high, 4, NULL) <<
+                               high_shift);
+    }
+
+/* was originally:
     st->st_rdev = (dev_t)((dev_t)iso_read_bb(pn->data.PN.high, 4, NULL) << 32)
                   | (dev_t)iso_read_bb(pn->data.PN.low, 4, NULL);
+*/
+
     return ISO_SUCCESS;
 }
+
+
+int read_aaip_AA(struct susp_sys_user_entry *sue,
+                 unsigned char **aa_string, size_t *aa_size, size_t *aa_len,
+                 size_t *prev_field, int *is_done, int flag)
+{
+     unsigned char *aapt;
+
+     if (*is_done) {
+         return ISO_WRONG_RR;
+     }
+
+     /* Eventually create or grow storage */
+     if (*aa_size == 0 || *aa_string == NULL) {
+         *aa_size = *aa_len + sue->len_sue[0];
+         *aa_string = calloc(*aa_size, 1);
+         *aa_len = 0;
+     } else if (*aa_len + sue->len_sue[0] > *aa_size) {
+         *aa_size += *aa_len + sue->len_sue[0];
+         *aa_string = realloc(*aa_string, *aa_size);
+     }
+     if (*aa_string == NULL)
+         return ISO_OUT_OF_MEM;
+
+     if (*aa_len > 0) {
+         /* Mark prev_field as being continued */
+         (*aa_string)[*prev_field + 4] = 1;
+     }
+
+     *prev_field = *aa_len;
+
+     /* Compose new SUSP header with signature aa[], cont == 0 */
+     aapt = *aa_string + *aa_len;
+     aapt[0] = 'A';
+     aapt[1] = 'A';
+     aapt[2] = sue->len_sue[0];
+     aapt[3] = 1;
+     aapt[4] = 0;
+     /* Append sue payload */
+     memcpy(aapt + 5, sue->data.AA.comps, sue->len_sue[0] - 5);
+
+     *aa_len += sue->len_sue[0];
+     *is_done = !(sue->data.AA.flags[0] & 1);
+
+     return ISO_SUCCESS;
+}
+
